﻿Imports System.IO

Public Class FormMain
    Private CurrentPath As String = ""
    Private FileInfoList As nMtp.FileInfo()
    ''' <summary>
    ''' デバイス一覧取得
    ''' </summary>
    Private Sub SetDevice()
        Dim device_list As String() = nMtp.Directory.GetDevices()
        For Each device As String In device_list
            Dim node As TreeNode = TreeFolder.Nodes.Add(device)
            node.ImageIndex = 0
            node.SelectedImageIndex = 0
            node.Tag = device
            SetFolder(node, device)
        Next
    End Sub
    ''' <summary>
    ''' フォルダツリー作成
    ''' </summary>
    ''' <param name="node">TreeViewノード</param>
    ''' <param name="path">デバイスパス名</param>
    Private Sub SetFolder(node As TreeNode, path As String)
        Try
            Dim dir As New nMtp.DirectoryInfo(path)
            Dim folder_list As nMtp.DirectoryInfo() = dir.GetDirectories()
            For Each folder As nMtp.DirectoryInfo In folder_list
                Dim sub_node As TreeNode = node.Nodes.Add(folder.Name)
                sub_node.ImageIndex = 1
                sub_node.SelectedImageIndex = 1
                sub_node.Tag = folder.FullName
                SetFolder(sub_node, folder.FullName)
            Next
        Catch ex As Exception
            Console.WriteLine(ex.Message)
        End Try
    End Sub
    ''' <summary>
    ''' ファイル一覧表示
    ''' </summary>
    ''' <param name="path">カレントパス</param>
    Private Sub GetFile(path As String)
        ListFile.VirtualListSize = 0
        Try
            Dim dir_info As New nMtp.DirectoryInfo(path)
            FileInfoList = dir_info.GetFiles()
            ListFile.VirtualListSize = FileInfoList.Length
        Catch ex As Exception
            Console.WriteLine(ex.Message)
        End Try
    End Sub

    Private Const itemSrc As Integer = 0
    Private Const itemDst As Integer = 1
    Private Const itemResult As Integer = 2
    Private Const itemError As Integer = 3
    Private Const itemMax As Integer = 4

    Private Sub FormMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        SetDevice()
    End Sub

    ''' <summary>
    ''' フォルダコピー
    ''' </summary>
    ''' <param name="flag">コピー設定</param>
    Private Sub FolderCopy(flag As Integer)
        Dim dialog As New nMtp.FolderBrowserDialog()
		dialog.Description = "コピーするフォルダを選択してください"
		If dialog.ShowDialog() = DialogResult.OK Then
            Dim dir_info As New nMtp.DirectoryInfo(CurrentPath)
            Try
                ' コピーファイル情報取得
                dir_info.CopyFirst(dialog.SelectPath, flag)
                ' 結果ダイアログ生成
                Dim form_result As New FormResult()
                Dim item As String() = New String(itemMax - 1) {}
                Dim count As Integer = 0
                ' 進捗表示ダイアログ生成
                Dim form_progress As New FormProgress()
                form_progress.Owner = Me
                form_progress.SetPos(count)
                form_progress.SetMax(dir_info.CopyCount)
                form_progress.Show()
                Application.DoEvents()
                ' 一つずつファイルコピー
                While dir_info.CopyNext()
                    ' 結果を追加
                    item(itemSrc) = dir_info.CopySourceFile
                    item(itemDst) = dir_info.CopyDestFile
                    form_progress.SetMessage(item(itemSrc), item(itemDst))
                    form_progress.SetPos(System.Threading.Interlocked.Increment(count))
                    Application.DoEvents()
                    If dir_info.CopyResult Then
                        item(itemResult) = "成功"
                        item(itemError) = ""
                    Else
                        item(itemResult) = "失敗"
                        item(itemError) = dir_info.CopyErrorString
                    End If
                    form_result.AddItem(item)
                    If form_progress.IsStop Then
                        Exit While
                    End If
                End While
                form_progress.Close()
                dir_info.CopyClose()
                ' 結果ダイアログ表示
                form_result.ShowDialog(Me)
            Catch ex As Exception
                MessageBox.Show("エラー:" + ex.Message)
            End Try
        End If
    End Sub

    Private Sub MenuItemFolderCopy_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemFolderCopy.Click
        ' フォルダコピー カレントフォルダのみ
        FolderCopy(0)
    End Sub

    Private Sub MenuItemSubFolderCopy_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemSubFolderCopy.Click
        ' フォルダコピー サブフォルダもコピー
        FolderCopy(nMtp.Directory.CopySubFolder)
    End Sub

    ''' <summary>
    ''' ファイルコピー
    ''' </summary>
    ''' <param name="flag">コピー設定</param>
    Private Sub FileCopy(flag As Integer)
        If ListFile.SelectedIndices.Count > 0 Then
            ' コピー先フォルダ選択
			Dim dialog As New nMtp.FolderBrowserDialog()
			dialog.Description = "コピーするフォルダを選択してください"
			If dialog.ShowDialog() = DialogResult.OK Then
                ' 結果ダイアログ生成
                Dim form_result As New FormResult()
                Dim item As String() = New String(itemMax - 1) {}
                Dim count As Integer = 0
                ' 進捗表示ダイアログ生成
                Dim form_progress As New FormProgress()
                form_progress.Owner = Me
                form_progress.SetPos(count)
                form_progress.SetMax(ListFile.SelectedIndices.Count)
                form_progress.Show()
                Application.DoEvents()
                For no As Integer = 0 To ListFile.SelectedIndices.Count - 1
                    Dim item_no As Integer = ListFile.SelectedIndices(no)
                    item(itemSrc) = FileInfoList(item_no).FullName
                    item(itemDst) = Path.Combine(dialog.SelectPath, FileInfoList(item_no).Name)
                    form_progress.SetMessage(item(itemSrc), item(itemDst))
                    form_progress.SetPos(System.Threading.Interlocked.Increment(count))
                    Application.DoEvents()
                    Try
                        ' 選択しているファイルをコピー
                        FileInfoList(item_no).CopyTo(item(itemDst), flag)

                        item(itemResult) = "成功"
                        item(itemError) = ""
                    Catch ex As Exception
                        item(itemResult) = "失敗"
                        item(itemError) = ex.Message
                    End Try
                    form_result.AddItem(item)
                    If form_progress.IsStop Then
                        Exit For
                    End If
                Next
                form_progress.Close()
                ' 結果ダイアログ表示
                form_result.ShowDialog(Me)
            End If
        End If
    End Sub

    Private Sub MenuItemFileCopy_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemFileCopy.Click
        ' ファイルコピー 上書きなし
        FileCopy(nMtp.FileInfo.CopyFailExists)
    End Sub

    Private Sub MenuItemOverwriteCopy_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemOverwriteCopy.Click
        ' ファイルコピー 上書きあり
        FileCopy(0)
    End Sub

    Private Sub MenuItemNewCopy_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemNewCopy.Click
        ' ファイルコピー 同名ファイル有りの場合新しければコピー
        FileCopy(nMtp.FileInfo.CopyNewer)
    End Sub
    ''' <summary>
    ''' 仮想リストビューアイテム設定
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub ListFile_RetrieveVirtualItem(sender As System.Object, e As System.Windows.Forms.RetrieveVirtualItemEventArgs) Handles ListFile.RetrieveVirtualItem
        Dim attr_string As String = If((FileInfoList(e.ItemIndex).Attributes <> 0), FileInfoList(e.ItemIndex).Attributes.ToString(), "")
        Dim date_string As String = If((FileInfoList(e.ItemIndex).IsValidLastWriteTime), FileInfoList(e.ItemIndex).LastWriteTime.ToString(), "")
        Dim item As String() = {FileInfoList(e.ItemIndex).Name, date_string, FileInfoList(e.ItemIndex).Length.ToString(), attr_string}
        e.Item = New ListViewItem(item)
    End Sub
    ''' <summary>
    ''' フォルダツリー選択後処理
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub TreeFolder_AfterSelect(sender As System.Object, e As System.Windows.Forms.TreeViewEventArgs) Handles TreeFolder.AfterSelect
        If TreeFolder.SelectedNode.Tag IsNot Nothing Then
            ' カレントパス設定
            CurrentPath = DirectCast(TreeFolder.SelectedNode.Tag, String)
            ' ファイル一覧表示
            GetFile(CurrentPath)
        End If
    End Sub

    Private Sub MenuItemDelete_Click(sender As System.Object, e As System.EventArgs) Handles MenuItemDelete.Click
        For no As Integer = 0 To ListFile.SelectedIndices.Count - 1
            Dim item_no As Integer = ListFile.SelectedIndices(no)
            Try
                ' ファイル削除
                FileInfoList(item_no).Delete()
                MessageBox.Show("削除成功:" + FileInfoList(item_no).Name)
            Catch ex As Exception
                MessageBox.Show("削除失敗:" + FileInfoList(item_no).Name + vbLf + ex.Message)
            End Try
        Next
        GetFile(CurrentPath)
    End Sub
End Class
