﻿Imports System.Text
Imports System.Runtime.InteropServices
Imports System.Collections
Imports System.Collections.Specialized
Imports System.Net.Sockets
Imports System.IO
Imports System.Runtime.Serialization
Imports System.Collections.Generic

Namespace nMtp
    ''' <summary>
    ''' ファイル情報構造体
    ''' </summary>
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Unicode)> _
    Public Structure WIN32_FIND_DATA
        ''' <summary>
        ''' 属性
        ''' </summary>
        Public dwFileAttributes As UInteger
        ''' <summary>
        ''' 作成日時
        ''' </summary>
        Public ftCreationTime As System.Runtime.InteropServices.ComTypes.FILETIME
        ''' <summary>
        ''' 最終アクセス日時
        ''' </summary>
        Public ftLastAccessTime As System.Runtime.InteropServices.ComTypes.FILETIME
        ''' <summary>
        ''' 最終更新日時
        ''' </summary>
        Public ftLastWriteTime As System.Runtime.InteropServices.ComTypes.FILETIME
        ''' <summary>
        ''' ファイルサイズ（上位32bit）
        ''' </summary>
        Public nFileSizeHigh As UInteger
        ''' <summary>
        ''' ファイルサイズ（下位32bit）
        ''' </summary>
        Public nFileSizeLow As UInteger
        ''' <summary>
        ''' 予約
        ''' </summary>
        Public dwReserved0 As UInteger
        ''' <summary>
        ''' 予約
        ''' </summary>
        Public dwReserved1 As UInteger
        ''' <summary>
        ''' ファイル名
        ''' </summary>
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=260)> _
        Public cFileName As String
        ''' <summary>
        ''' <para>8.3形式のファイル名</para>
        ''' <para>MTPデバイス上のファイルまたはフォルダの場合 Object ID</para>
        ''' </summary>
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=14)> _
        Public cAlternateFileName As String
    End Structure

    ''' <summary>
    ''' ファイルシステム基本クラス
    ''' </summary>
    <SerializableAttribute()> _
    <ComVisibleAttribute(True)> _
    Public Class FileSystemInfo
        Inherits System.IO.FileSystemInfo
        Protected _read_flag As Boolean = False
        Protected _attributes As FileAttributes = CType(-1, FileAttributes)
        Protected _creation_time As New DateTime()
        Protected _last_access_time As New DateTime()
        Protected _last_write_time As New DateTime()
        Protected _length As Long = 0

        Protected MAX_PATH As Integer = 260
        Protected Const FILE_ATTRIBUTE_DIRECTORY As UInteger = &H10

        ''' <summary>
        ''' <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
            OriginalPath = ""
            FullPath = ""
            _read_flag = False
        End Sub

        ''' <summary>
        ''' <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">ファイルまたはフォルダ名</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Sub New(path_name As String)
            Directory.CheckInvalidPathName(path_name, False)

            OriginalPath = path_name
            If Not NMtpCheckDevicePathName(path_name) Then
                FullPath = Path.GetFullPath(path_name)
            Else
                FullPath = path_name
            End If
            _read_flag = False
        End Sub

        ''' <summary>
        ''' <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">ファイルまたはフォルダ名</param>
        ''' <param name="find_data">ファイルまたはフォルダの情報</param>
        Public Sub New(path_name As String, find_data As WIN32_FIND_DATA)
            Directory.CheckInvalidPathName(path_name, False)
            _attributes = CType(find_data.dwFileAttributes, FileAttributes)
            _creation_time = DateTime.FromFileTime((CLng(find_data.ftCreationTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftCreationTime.dwLowDateTime.ToString("X"), 16))
            _last_access_time = DateTime.FromFileTime((CLng(find_data.ftLastAccessTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastAccessTime.dwLowDateTime.ToString("X"), 16))
            _last_write_time = DateTime.FromFileTime((CLng(find_data.ftLastWriteTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastWriteTime.dwLowDateTime.ToString("X"), 16))
            _length = (CLng(find_data.nFileSizeHigh) << 32) + Convert.ToUInt32(find_data.nFileSizeLow.ToString("X"), 16)
            OriginalPath = Path.Combine(path_name, find_data.cFileName)
            FullPath = OriginalPath
            _read_flag = True
        End Sub

        Protected Sub GetStatus()
            If Not _read_flag Then
                Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()

                If NMtpGetStatus(FullPath, find_data) Then
                    _read_flag = True
                    _attributes = CType(find_data.dwFileAttributes, FileAttributes)
                    _creation_time = DateTime.FromFileTime((CLng(find_data.ftCreationTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftCreationTime.dwLowDateTime.ToString("X"), 16))
                    _last_access_time = DateTime.FromFileTime((CLng(find_data.ftLastAccessTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastAccessTime.dwLowDateTime.ToString("X"), 16))
                    _last_write_time = DateTime.FromFileTime((CLng(find_data.ftLastWriteTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastWriteTime.dwLowDateTime.ToString("X"), 16))
                    _length = (CLng(find_data.nFileSizeHigh) << 32) + Convert.ToUInt32(find_data.nFileSizeLow.ToString("X"), 16)
                End If
            End If
        End Sub

        Protected Shared Function CheckDirectory(attr As UInteger) As Boolean
            If (attr And FILE_ATTRIBUTE_DIRECTORY) = FILE_ATTRIBUTE_DIRECTORY Then
                Return True
            End If
            Return False
        End Function

        Protected Shared Function IsDots(name As String) As Boolean
            If name = "." OrElse name = ".." Then
                Return True
            End If
            Return False
        End Function

        ''' <summary>
        ''' ファイルを削除
        ''' </summary>
        Public Overrides Sub Delete()
            If IsDirectory Then
                Directory.Delete(FullPath)
            Else
                File.Delete(FullPath)
            End If
        End Sub

        ''' <summary>
        ''' ファイル情報を更新
        ''' </summary>
        Public Shadows Sub Refresh()
            _read_flag = False
            GetStatus()
        End Sub

        ''' <summary>
        ''' ファイルが存在するかを示す値を取得
        ''' </summary>
        Public Overrides ReadOnly Property Exists() As Boolean
            Get
                If IsDirectory Then
                    Return Directory.Exists(FullPath)
                Else
                    Return File.Exists(FullPath)
                End If
            End Get
        End Property

        ''' <summary>
        ''' ファイルの絶対パス名を取得
        ''' </summary>
        Public Overrides ReadOnly Property FullName() As String
            Get
                Return FullPath
            End Get
        End Property

        ''' <summary>
        ''' ファイルの名前を取得
        ''' </summary>
        Public Overrides ReadOnly Property Name() As String
            Get
                Return Path.GetFileName(FullName)
            End Get
        End Property

        ''' <summary>
        ''' MTP デバイス上のファイルまたはフォルダかどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsMtpDevice() As Boolean
            Get
                Return NMtpCheckDevicePathName(FullPath)
            End Get
        End Property

        ''' <summary>
        ''' フォルダかどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsDirectory() As Boolean
            Get
                GetStatus()
                If (_attributes And FileAttributes.Directory) = FileAttributes.Directory Then
                    Return True
                End If
                Return False
            End Get
        End Property

        ''' <summary>
        ''' <para>ファイルまたはフォルダの属性の取得または設定</para>
        ''' <para>フォルダまたは MTP デバイス上のファイルの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property Attributes() As FileAttributes
            Get
                GetStatus()
                Return _attributes
            End Get
            Set(value As FileAttributes)
                If Not IsMtpDevice Then
                    If Not IsDirectory Then
                        _attributes = value
                        System.IO.File.SetAttributes(FullPath, _attributes)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' ファイルまたはフォルダの作成日時が有効かどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsValidCreationTime() As Boolean
            Get
                If _creation_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) <> 0 Then
                    Return True
                End If
                Return False
            End Get
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの作成日時(世界協定時刻)の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property CreationTimeUtc() As DateTime
            Get
                GetStatus()
                Return _creation_time
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    _creation_time = Value
                    If IsDirectory Then
                        System.IO.Directory.SetCreationTimeUtc(FullPath, _creation_time)
                    Else
                        System.IO.File.SetCreationTimeUtc(FullPath, _creation_time)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの作成日時の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property CreationTime() As DateTime
            Get
                GetStatus()
                Return _creation_time.ToLocalTime()
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    If IsDirectory Then
                        _creation_time = Value.ToUniversalTime()
                        System.IO.Directory.SetCreationTimeUtc(FullPath, _creation_time)
                    Else
                        System.IO.File.SetCreationTimeUtc(FullPath, _creation_time)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' ファイルまたはフォルダの最終アクセス日時が有効かどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsValidLastAccessTime() As Boolean
            Get
                If _last_access_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) <> 0 Then
                    Return True
                End If
                Return False
            End Get
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの最終アクセス日時(世界協定時刻)の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property LastAccessTimeUtc() As DateTime
            Get
                GetStatus()
                Return _last_access_time
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    _last_access_time = Value
                    If IsDirectory Then
                        System.IO.Directory.SetLastAccessTimeUtc(FullPath, _last_access_time)
                    Else
                        System.IO.File.SetLastAccessTimeUtc(FullPath, _last_access_time)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの最終アクセス日時の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property LastAccessTime() As DateTime
            Get
                GetStatus()
                Return _last_access_time.ToLocalTime()
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    _last_access_time = Value.ToUniversalTime()
                    If IsDirectory Then
                        System.IO.Directory.SetLastAccessTimeUtc(FullPath, _last_access_time)
                    Else
                        System.IO.File.SetLastAccessTimeUtc(FullPath, _last_access_time)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' ファイルまたはフォルダの最終書込日時が有効かどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsValidLastWriteTime() As Boolean
            Get
                If _last_write_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) <> 0 Then
                    Return True
                End If
                Return False
            End Get
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの最終書込日時(世界協定時刻)の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property LastWriteTimeUtc() As DateTime
            Get
                GetStatus()
                Return _last_write_time
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    _last_write_time = Value
                    If IsDirectory Then
                        System.IO.Directory.SetLastWriteTimeUtc(FullPath, _last_write_time)
                    Else
                        System.IO.File.SetLastWriteTimeUtc(FullPath, _last_write_time)
                    End If
                End If
            End Set
        End Property
        ''' <summary>
        ''' <para>ファイルまたはフォルダの最終書込日時の取得または設定</para>
        ''' <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
        ''' </summary>
        Public Overridable Shadows Property LastWriteTime() As DateTime
            Get
                GetStatus()
                Return _last_write_time.ToLocalTime()
            End Get
            Set(value As DateTime)
                If Not IsMtpDevice Then
                    _last_write_time = Value.ToUniversalTime()
                    If IsDirectory Then
                        System.IO.Directory.SetLastWriteTime(FullPath, _last_write_time)
                    Else
                        System.IO.File.SetLastWriteTimeUtc(FullPath, _last_write_time)
                    End If
                End If
            End Set
        End Property

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCheckDevicePathName", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpCheckDevicePathName(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpGetStatus", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpGetStatus(path_name As String, ByRef find_data As WIN32_FIND_DATA) As Boolean
        End Function
    End Class

    ''' <summary>
    ''' フォルダ操作クラス
    ''' </summary>
    <SerializableAttribute()> _
    <ComVisibleAttribute(True)> _
    Public Class DirectoryInfo
        Inherits FileSystemInfo
        ''' <summary>
        ''' <para>コピー先にファイルがあれば失敗</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyFailExists As Integer = &H1
        ''' <summary>
        ''' <para>新しいファイルのみコピー</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyNewer As Integer = &H10
        ''' <summary>
        ''' <para>サブフォルダの中身もコピー</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopySubFolder As Integer = &H100
        ''' <summary>
        ''' <para>一度にコピー</para>
        ''' <para>CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyAll As Integer = &H1000
        ''' <summary>
        ''' <para>一度にコピーして結果も保存しない</para>
        ''' </summary>
        Public Const CopyAllNoResult As Integer = &H2000

        Private Const INVALID_HANDLE_VALUE As Integer = -1
        Private Const MAX_DIR_PATH As Integer = 248

        Private _handle As IntPtr = IntPtr.Zero
        Private _source_file As String = ""
        Private _dest_file As String = ""
        Private _copy_result As Integer = 0
        Private _copy_error As Integer = 0

        Private Const ERROR_FILE_NOT_FOUND As Integer = 2
        Private Const ERROR_PATH_NOT_FOUND As Integer = 3
        Private Const ERROR_FILE_EXISTS As Integer = 80

        ''' <summary>
        ''' <c>DirectoryInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">フォルダ名</param>
        Public Sub New(path_name As String)
            MyBase.New(path_name)
        End Sub

        ''' <summary>
        ''' <c>DirectoryInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">フォルダ名</param>
        ''' <param name="find_data">フォルダの情報</param>
        Public Sub New(path_name As String, find_data As WIN32_FIND_DATA)
            MyBase.New(path_name, find_data)
        End Sub

        ''' <summary>
        ''' フォルダ作成
        ''' </summary>
        '''	<exception cref="IOException">
        '''	フォルダを作成できない
        ''' </exception>
        Public Sub Create()
            Directory.CreateDirectory(FullPath)
        End Sub

        ''' <summary>
        ''' サブフォルダ作成
        ''' </summary>
        ''' <param name="path_name">サブフォルダ名</param>
        '''	<exception cref="IOException">
        '''	サブフォルダを作成できない
        ''' </exception>
        Public Sub CreateSubdirectory(path_name As String)
            Directory.CreateDirectory(Path.Combine(FullPath, path_name))
        End Sub

        ''' <summary>
        ''' フォルダ削除
        ''' </summary>
        '''	<exception cref="IOException">
        '''	サブフォルダを削除できない
        ''' </exception>
        Public Overrides Sub Delete()
            Directory.Delete(FullPath)
        End Sub

        ''' <summary>
        ''' フォルダ削除
        ''' </summary>
        ''' <param name="recursive">true でサブフォルダ、ファイルも含めて削除</param>
        '''	<exception cref="IOException">
        '''	サブフォルダを削除できない
        ''' </exception>
        Public Overloads Sub Delete(recursive As Boolean)
            Directory.Delete(FullPath, recursive)
        End Sub

        Protected Sub GetDirectoryList(path_name As String, searchPattern As String, ByRef list As List(Of DirectoryInfo))
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            h = NMtpFindFirstFile(Path.Combine(path_name, "*"), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If CheckDirectory(find_data.dwFileAttributes) Then
                        Dim name As String = find_data.cFileName.ToString()
                        If Not IsDots(name) Then
                            If NMtpWildcardMatch(searchPattern, name) Then
                                list.Add(New DirectoryInfo(path_name, find_data))
                            End If
                            GetDirectoryList(Path.Combine(path_name, name), searchPattern, list)
                        End If
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
        End Sub

        ''' <summary>
        ''' 指定されたフォルダ内のサブフォルダを取得
        ''' </summary>
        ''' <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
        Public Function GetDirectories() As DirectoryInfo()
            Return GetDirectories("*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のサブフォルダを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
        Public Function GetDirectories(searchPattern As String) As DirectoryInfo()
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of DirectoryInfo)()
            h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If CheckDirectory(find_data.dwFileAttributes) Then
                        list.Add(New DirectoryInfo(FullPath, find_data))
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のサブフォルダを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <param name="searchOptionType">検索オプション</param>
        ''' <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
        Public Function GetDirectories(searchPattern As String, searchOptionType As SearchOption) As DirectoryInfo()
            If searchOptionType = SearchOption.AllDirectories Then
                Dim list As New List(Of DirectoryInfo)()
                GetDirectoryList(FullPath, searchPattern, list)
                Return list.ToArray()
            End If
            Return GetDirectories(searchPattern)
        End Function

        Protected Sub GetFileList(path_name As String, searchPattern As String, ByRef list As List(Of FileInfo))
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            h = NMtpFindFirstFile(Path.Combine(path_name, "*"), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    Dim name As String = find_data.cFileName.ToString()
                    If CheckDirectory(find_data.dwFileAttributes) Then
                        If Not IsDots(name) Then
                            GetFileList(Path.Combine(path_name, name), searchPattern, list)
                        End If
                    Else
                        If NMtpWildcardMatch(searchPattern, name) Then
                            list.Add(New FileInfo(path_name, find_data))
                        End If
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
        End Sub

        ''' <summary>
        ''' 指定されたフォルダ内のファイルを取得
        ''' </summary>
        ''' <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
        Public Function GetFiles() As FileInfo()
            Return GetFiles("*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
        Public Function GetFiles(searchPattern As String) As FileInfo()
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of FileInfo)()

            h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If Not CheckDirectory(find_data.dwFileAttributes) Then
                        list.Add(New FileInfo(FullPath, find_data))
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <param name="searchOptionType">検索オプション</param>
        ''' <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
        Public Function GetFiles(searchPattern As String, searchOptionType As SearchOption) As FileInfo()
            If searchOptionType = SearchOption.AllDirectories Then
                Dim list As New List(Of FileInfo)()
                GetFileList(FullPath, searchPattern, list)
                Return list.ToArray()
            End If
            Return GetFiles(searchPattern)
        End Function

        Protected Sub GetFileSystemList(path_name As String, searchPattern As String, ByRef list As List(Of FileSystemInfo))
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            h = NMtpFindFirstFile(Path.Combine(path_name, "*"), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    Dim name As String = find_data.cFileName.ToString()
                    If CheckDirectory(find_data.dwFileAttributes) Then
                        If Not IsDots(name) Then
                            If NMtpWildcardMatch(searchPattern, name) Then
                                list.Add(New FileSystemInfo(path_name, find_data))
                            End If
                            GetFileSystemList(Path.Combine(path_name, name), searchPattern, list)
                        End If
                    Else
                        If NMtpWildcardMatch(searchPattern, name) Then
                            list.Add(New FileSystemInfo(path_name, find_data))
                        End If
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
        End Sub

        ''' <summary>
        ''' 指定されたフォルダ内のファイルおよびサブフォルダを取得
        ''' </summary>
        ''' <returns>指定したパス内のサブフォルダのファイルおよびサブフォルダ名の配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
        Public Function GetFileSystemInfos() As FileSystemInfo()
            Return GetFileSystemInfos("*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルおよびサブフォルダを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のサブフォルダのファイルおよびサブフォルダ名の配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
        Public Function GetFileSystemInfos(searchPattern As String) As FileSystemInfo()
            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of FileSystemInfo)()

            h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If Not CheckDirectory(find_data.dwFileAttributes) Then
                        list.Add(New FileSystemInfo(Path.Combine(FullPath, find_data.cFileName.ToString())))
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルおよびサブフォルダを取得
        ''' </summary>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <param name="searchOptionType">検索オプション</param>
        ''' <returns>指定したパス内のファイルおよびサブフォルダの配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
        Public Function GetFileSystemInfos(searchPattern As String, searchOptionType As SearchOption) As FileSystemInfo()
            If searchOptionType = SearchOption.AllDirectories Then
                Dim list As New List(Of FileSystemInfo)()
                GetFileSystemList(FullPath, searchPattern, list)
                Return list.ToArray()
            End If
            Return GetFiles(searchPattern)
        End Function

        ''' <summary>
        ''' フォルダファイルコピー
        ''' </summary>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <returns>true で成功</returns>
        Public Function Copy(destFolderName As String) As Boolean
            If NMtpCopyFolderFirst(FullPath, destFolderName, CopyAllNoResult) = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' フォルダファイルコピー
        ''' </summary>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー設定</param>
        ''' <returns>true で成功</returns>
        Public Function Copy(destFolderName As String, flag As Integer) As Boolean
            If NMtpCopyFolderFirst(FullPath, destFolderName, flag Or CopyAllNoResult) = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' フォルダファイルコピー
        ''' </summary>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー設定</param>
        ''' <param name="copyPattern">ファイル指定文字列</param>
        ''' <returns>true で成功</returns>
        Public Function Copy(destFolderName As String, flag As Integer, copyPattern As String) As Boolean
            If NMtpCopyFolderFirst(Path.Combine(FullPath, copyPattern), destFolderName, flag Or CopyAllNoResult) = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピーファイルの取得</para>
        ''' </summary>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー動作設定</param>
        ''' <exception cref="ArgumentNullException">
        '''	フォルダ が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	フォルダ がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Sub CopyFirst(destFolderName As String, flag As Integer)
            Directory.CheckInvalidPathName(destFolderName, False)

            _handle = NMtpCopyFolderFirst(FullPath, destFolderName, flag)
            If _handle = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                _handle = IntPtr.Zero
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピーファイルの取得</para>
        ''' </summary>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー動作設定</param>
        ''' <param name="copyPattern">ファイル指定文字列</param>
        ''' <exception cref="ArgumentNullException">
        '''	フォルダ が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	フォルダ がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Sub CopyFirst(destFolderName As String, flag As Integer, copyPattern As String)
            Directory.CheckInvalidPathName(destFolderName, False)

            _handle = NMtpCopyFolderFirst(Path.Combine(FullPath, copyPattern), destFolderName, flag)
            If _handle = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                _handle = IntPtr.Zero
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピー実行</para>
        ''' </summary>
        ''' <returns>true でコピーファイル有</returns>
        Public Function CopyNext() As Boolean
            _source_file = ""
            _dest_file = ""
            _copy_result = 0
            _copy_error = 0
            If _handle <> IntPtr.Zero Then
                Dim src As New StringBuilder(MAX_PATH)
                Dim dst As New StringBuilder(MAX_PATH)
                If NMtpCopyFolderNext(_handle, src, MAX_PATH, dst, MAX_PATH, _copy_result, _
                 _copy_error) Then
                    _source_file = src.ToString()
                    _dest_file = dst.ToString()
                    Return True
                End If
            End If
            Return False
        End Function

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>終了</para>
        ''' </summary>
        Public Sub CopyClose()
            If _handle <> IntPtr.Zero Then
                NMtpCopyFolderClose(_handle)
                _handle = IntPtr.Zero
            End If
        End Sub

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>エラー番号取得</para>
        ''' </summary>
        Public ReadOnly Property CopyError() As Integer
            Get
                Return _copy_error
            End Get
        End Property

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピー元ファイル取得</para>
        ''' </summary>
        Public ReadOnly Property CopySourceFile() As String
            Get
                Return _source_file
            End Get
        End Property

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピー結果</para>
        ''' </summary>
        Public ReadOnly Property CopyResult() As Boolean
            Get
                If _copy_result <> 0 Then
                    Return True
                End If
                Return False
            End Get
        End Property

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピー先ファイル取得</para>
        ''' </summary>
        Public ReadOnly Property CopyDestFile() As String
            Get
                Return _dest_file
            End Get
        End Property
        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>エラー文字列取得</para>
        ''' </summary>
        Public ReadOnly Property CopyErrorString() As String
            Get
                If _copy_error = ERROR_PATH_NOT_FOUND Then
                    Return "Path not found"
                ElseIf _copy_error = ERROR_FILE_NOT_FOUND Then
                    Return "File not found"
                ElseIf _copy_error = ERROR_FILE_EXISTS Then
                    Return "File Exists"
                End If
                Return _copy_error.ToString()
            End Get
        End Property

        ''' <summary>
        ''' フォルダファイルコピーのファイル数
        ''' </summary>
        Public ReadOnly Property CopyCount() As Integer
            Get
                If _handle = IntPtr.Zero Then
                    Return 0
                Else
                    Return NMtpCopyFolderCount(_handle)
                End If
            End Get
        End Property

        ''' <summary>
        ''' 親フォルダ名の取得
        ''' </summary>
        Public ReadOnly Property Parent() As String
            Get
                Return Path.GetDirectoryName(FullPath)
            End Get
        End Property
        ''' <summary>
        ''' ルートパスの取得
        ''' </summary>
        Public ReadOnly Property Root() As String
            Get
                Return Path.GetPathRoot(FullPath)
            End Get
        End Property

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindFirstFile", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpFindFirstFile(file_name As String, ByRef find_data As WIN32_FIND_DATA) As IntPtr
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindNextFile", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpFindNextFile(handle As IntPtr, ByRef find_data As WIN32_FIND_DATA) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindClose")> _
        Protected Shared Function NMtpFindClose(handle As IntPtr) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderFirst", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpCopyFolderFirst(src As String, dst As String, flag As Integer) As IntPtr
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderCount")> _
        Protected Shared Function NMtpCopyFolderCount(handle As IntPtr) As Integer
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderNext", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpCopyFolderNext(handle As IntPtr, src_name As StringBuilder, src_length As Integer, dst_name As StringBuilder, dst_length As Integer, ByRef result As Integer, _
   ByRef error_code As Integer) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderClose")> _
        Protected Shared Function NMtpCopyFolderClose(handle As IntPtr) As Integer
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpWildcardMatch", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpWildcardMatch(pattern As String, name As String) As Boolean
        End Function
    End Class

    ''' <summary>
    ''' ファイル操作クラス
    ''' </summary>
    <SerializableAttribute()> _
    <ComVisibleAttribute(True)> _
    Public Class FileInfo
        Inherits FileSystemInfo
        ''' <summary>
        ''' <para>コピー先にファイルがあれば失敗</para>
        ''' <para>CopyTo の flag で指定</para>
        ''' </summary>
        Public Const CopyFailExists As Integer = &H1
        ''' <summary>
        ''' <para>新しいファイルのみコピー</para>
        ''' <para>CopyTo の flag で指定</para>
        ''' </summary>
        Public Const CopyNewer As Integer = &H10

        Private Const INVALID_HANDLE_VALUE As Integer = -1
        Private Const MAX_DIR_PATH As Integer = 248

        ''' <summary>
        ''' <c>FileInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">ファイル名</param>
        Public Sub New(path_name As String)
            MyBase.New(path_name)
        End Sub

        ''' <summary>
        ''' <c>FileInfo</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="path_name">ファイル名</param>
        ''' <param name="find_data">ファイルの情報</param>
        Public Sub New(path_name As String, find_data As WIN32_FIND_DATA)
            MyBase.New(path_name, find_data)
        End Sub

        ''' <summary>
        ''' ファイルの削除
        ''' </summary>
        '''	<exception cref="IOException">
        '''	ファイルを削除できない
        ''' </exception>
        Public Overrides Sub Delete()
            File.Delete(FullPath)
        End Sub

        ''' <summary>
        ''' ファイルのコピー
        ''' </summary>
        ''' <param name="destFileName">コピー先のファイルパス</param>
        '''	<exception cref="IOException">
        '''	ファイルをコピーできない
        ''' </exception>
        Public Sub CopyTo(destFileName As String)
            CopyTo(destFileName, False)
        End Sub

        ''' <summary>
        ''' ファイルの移動
        ''' </summary>
        ''' <param name="destFileName">移動先のファイルパス</param>
        '''	<exception cref="IOException">
        '''	ファイルを移動できない
        ''' </exception>
        Public Sub MoveTo(destFileName As String)
            File.Move(FullPath, destFileName)
        End Sub

        ''' <summary>
        ''' ファイルのコピー
        ''' </summary>
        ''' <param name="destFileName">コピー先のファイルパス</param>
        ''' <param name="overwrite">true で上書</param>
        '''	<exception cref="IOException">
        '''	ファイルをコピーできない
        ''' </exception>
        Public Sub CopyTo(destFileName As String, overwrite As Boolean)
            File.Copy(FullPath, destFileName, overwrite)
        End Sub

        ''' <summary>
        ''' ファイルのコピー
        ''' </summary>
        ''' <param name="destFileName">コピー先のファイルパス</param>
        ''' <param name="flag">コピー設定</param>
        '''	<exception cref="IOException">
        '''	ファイルをコピーできない
        ''' </exception>
        Public Sub CopyTo(destFileName As String, flag As Integer)
            File.Copy(FullPath, destFileName, flag)
        End Sub

        ''' <summary>
        ''' ファイルが読出し専用かどうかを示す値を取得
        ''' </summary>
        Public ReadOnly Property IsReadOnly() As Boolean
            Get
                GetStatus()
                If (_attributes And FileAttributes.ReadOnly) = FileAttributes.ReadOnly Then
                    Return True
                End If
                Return False
            End Get
        End Property


        ''' <summary>
        ''' 親フォルダのインスタンスを取得
        ''' </summary>
        Public ReadOnly Property Directory() As DirectoryInfo
            Get
                Return New DirectoryInfo(Path.GetDirectoryName(FullPath))
            End Get
        End Property

        ''' <summary>
        ''' 親フォルダ名を取得
        ''' </summary>
        Public ReadOnly Property DirectoryName() As String
            Get
                Return Path.GetDirectoryName(FullPath)
            End Get
        End Property

        ''' <summary>
        ''' ファイルサイズを取得
        ''' </summary>
        Public ReadOnly Property Length() As Long
            Get
                GetStatus()
                Return _length
            End Get
        End Property

    End Class

    ''' <summary>
    ''' フォルダ操作クラス
    ''' </summary>
    Public Class Directory
        ''' <summary>
        ''' <para>コピー先にファイルがあれば失敗</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyFailExists As Integer = &H1
        ''' <summary>
        ''' <para>新しいファイルのみコピー</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyNewer As Integer = &H10
        ''' <summary>
        ''' <para>サブフォルダの中身もコピー</para>
        ''' <para>Copy/CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopySubFolder As Integer = &H100
        ''' <summary>
        ''' <para>一度にコピー</para>
        ''' <para>CopyFirst の flag で指定</para>
        ''' </summary>
        Public Const CopyAll As Integer = &H1000
        ''' <summary>
        ''' <para>一度にコピーして結果も保存しない</para>
        ''' </summary>
        Public Const CopyAllNoResult As Integer = &H2000

        Private Const ERROR_FILE_NOT_FOUND As Integer = 2
        Private Const ERROR_PATH_NOT_FOUND As Integer = 3
        Private Const ERROR_ACCESS_DENIED As Integer = 5
        Private Const ERROR_FILE_EXISTS As Integer = 80
        Private Const ERROR_CANNOT_MAKE As Integer = 82

        Private Const INVALID_HANDLE_VALUE As Integer = -1
        Private Const MAX_OBJECT_ID As Integer = 14
        Private Const MAX_PATH As Integer = 260
        Private Const MAX_DIR_PATH As Integer = 248

        Private _handle As IntPtr = IntPtr.Zero

        Private Const FILE_ATTRIBUTE_DIRECTORY As UInteger = &H10

        ''' <summary>
        ''' <c>Directory</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
        End Sub

        ''' <summary>
        ''' MTP 接続デバイス一覧取得
        ''' </summary>
        ''' <returns>デバイス名一覧</returns>
        Public Shared Function GetDevices() As String()
            Dim h As IntPtr
            Dim name As StringBuilder
            name = New StringBuilder(MAX_PATH)
            Dim list As New List(Of String)()

            h = NMtpFindDeviceFirst()
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                While NMtpFindDeviceNext(h, name, MAX_PATH)
                    list.Add(name.ToString())
                End While
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' パス名のチェック
        ''' </summary>
        ''' <param name="path_name">パス名</param>
        ''' <param name="wildcard">ファイル名にワイルドカードが入る</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Sub CheckInvalidPathName(path_name As String, wildcard As Boolean)
            If path_name Is Nothing Then
                Throw New ArgumentNullException()
            End If
            If path_name.Length >= MAX_DIR_PATH Then
                Throw New PathTooLongException()
            End If
            Dim space As String = path_name.Trim()
            If space = String.Empty Then
                Throw New ArgumentNullException()
            End If
            Dim invalid_chars As String = """<>|/" & vbNullChar & ChrW(1) & ChrW(2) & ChrW(3) & ChrW(4) & ChrW(5) & ChrW(6) & ChrW(7) & vbBack & vbTab & vbLf & vbVerticalTab & vbFormFeed & vbCr & ChrW(14) & ChrW(15) & ChrW(16) & ChrW(17) & ChrW(18) & ChrW(19) & ChrW(20) & ChrW(21) & ChrW(22) & ChrW(23) & ChrW(24) & ChrW(25) & ChrW(26) & ChrW(27) & ChrW(28) & ChrW(29) & ChrW(30) & ChrW(31)
            If path_name.IndexOfAny(invalid_chars.ToCharArray()) >= 0 Then
                Throw New ArgumentException()
            End If
            Dim dirs As String = Path.GetDirectoryName(path_name)
            invalid_chars = "*?"
            If dirs IsNot Nothing Then
                If dirs.IndexOfAny(invalid_chars.ToCharArray()) >= 0 Then
                    Throw New ArgumentException()
                End If
            End If
            Dim name As String = Path.GetFileName(path_name)
            If name IsNot Nothing Then
                If wildcard Then
                    invalid_chars = ":\"
                Else
                    invalid_chars += ":\"
                End If
                If name.IndexOfAny(invalid_chars.ToCharArray()) >= 0 Then
                    Throw New ArgumentException()
                End If
            End If
        End Sub

        ''' <summary>
        ''' 指定されたフォルダ内のサブフォルダの名前を取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <returns>指定したパス内のサブフォルダ名の配列。フォルダが見つからない場合は空の配列</returns>
        Public Shared Function GetDirectories(path_name As String) As String()
            Return GetDirectories(path_name, "*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のサブフォルダの名前を取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のサブフォルダ名の配列。フォルダが見つからない場合は空の配列</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Function GetDirectories(path_name As String, searchPattern As String) As String()
            CheckInvalidPathName(path_name, False)

            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of String)()

            h = NMtpFindFirstFile(path.Combine(path_name, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = FILE_ATTRIBUTE_DIRECTORY Then
                        list.Add(path.Combine(path_name, find_data.cFileName.ToString()))
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイル名を取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <returns>指定したパス内のサブフォルダのファイル名の配列。ファイルが見つからない場合は空の配列</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Function GetFiles(path_name As String) As String()
            Return GetFiles(path_name, "*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイル名を取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のサブフォルダのファイル名の配列。ファイルが見つからない場合は空の配列</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Function GetFiles(path_name As String, searchPattern As String) As String()
            CheckInvalidPathName(path_name, False)

            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of String)()

            h = NMtpFindFirstFile(path.Combine(path_name, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = 0 Then
                        list.Add(path.Combine(path_name, find_data.cFileName.ToString()))
                    End If
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルとフォルダの名前の取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <returns>指定したパス内のファイルとサブフォルダの名前の配列。ファイルが見つからない場合は空の配列</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Function GetFileSystemEntries(path_name As String) As String()
            Return GetFileSystemEntries(path_name, "*")
        End Function

        ''' <summary>
        ''' 指定されたフォルダ内のファイルとフォルダの名前の取得
        ''' </summary>
        ''' <param name="path_name">検索するフォルダのパス</param>
        ''' <param name="searchPattern">検索文字列</param>
        ''' <returns>指定したパス内のファイルとサブフォルダの名前の配列。ファイルが見つからない場合は空の配列</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        Public Shared Function GetFileSystemEntries(path_name As String, searchPattern As String) As String()
            CheckInvalidPathName(path_name, False)

            Dim h As IntPtr
            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim list As New List(Of String)()

            h = NMtpFindFirstFile(path.Combine(path_name, searchPattern), find_data)
            If h <> CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Do
                    list.Add(path.Combine(path_name, find_data.cFileName.ToString()))
                Loop While NMtpFindNextFile(h, find_data)
                NMtpFindClose(h)
            End If
            Return list.ToArray()
        End Function

        ''' <summary>
        ''' フォルダの作成
        ''' </summary>
        ''' <param name="path_name">作成するフォルダ</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダが作成できない
        ''' </exception>
        Public Shared Sub CreateDirectory(path_name As String)
            CheckInvalidPathName(path_name, False)

            If Not NMtpCreateFolder(path_name) Then
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' フォルダの削除
        ''' </summary>
        ''' <param name="path_name">削除するフォルダ</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダが作成できない
        ''' </exception>
        Public Shared Sub Delete(path_name As String)
            CheckInvalidPathName(path_name, False)

            If Not NMtpRemoveFolder(path_name) Then
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' フォルダの削除
        ''' </summary>
        ''' <param name="path_name">削除するフォルダ</param>
        ''' <param name="recursive">true で指定されたフォルダのファイル、サブフォルダも削除</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダが作成できない
        ''' </exception>
        Public Shared Sub Delete(path_name As String, recursive As Boolean)
            CheckInvalidPathName(path_name, False)

            If recursive Then
                If Not NMtpDeleteFolder(path_name) Then
                    Dim error_code As Integer = Marshal.GetLastWin32Error()
                    Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                    If error_code = ERROR_PATH_NOT_FOUND Then
                        Throw New DirectoryNotFoundException(message)
                    Else
                        Throw New IOException(message)
                    End If
                End If
            Else
                If Not NMtpRemoveFolder(path_name) Then
                    Dim error_code As Integer = Marshal.GetLastWin32Error()
                    Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                    If error_code = ERROR_PATH_NOT_FOUND Then
                        Throw New DirectoryNotFoundException(message)
                    Else
                        Throw New IOException(message)
                    End If
                End If
            End If
        End Sub

        ''' <summary>
        ''' フォルダの存在確認
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>true で存在</returns>
        Public Shared Function Exists(path_name As String) As Boolean
            CheckInvalidPathName(path_name, False)

            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()

            If NMtpGetStatus(path_name, find_data) Then
                If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = FILE_ATTRIBUTE_DIRECTORY Then
                    Return True
                End If
            End If
            Return False
        End Function

        Private Const timeCreation As Integer = 0
        Private Const timeLastWrite As Integer = 1
        Private Const timeLastAccess As Integer = 2

        Private Shared Function GetTime(path_name As String, time_type As Integer, local_flag As Boolean) As DateTime
            CheckInvalidPathName(path_name, False)

            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim date_time As New DateTime()

            If NMtpGetStatus(path_name, find_data) Then
                If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = FILE_ATTRIBUTE_DIRECTORY Then
                    If time_type = timeLastWrite Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftLastWriteTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastWriteTime.dwLowDateTime.ToString("X"), 16))
                    ElseIf time_type = timeCreation Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftCreationTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftCreationTime.dwLowDateTime.ToString("X"), 16))
                    ElseIf time_type = timeLastAccess Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftLastAccessTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastAccessTime.dwLowDateTime.ToString("X"), 16))
                    End If
                Else
                    Dim message As String = New System.ComponentModel.Win32Exception(ERROR_PATH_NOT_FOUND).Message
                    Throw New DirectoryNotFoundException(message)
                End If
            Else
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
            If local_flag Then
                date_time.ToLocalTime()
            End If
            Return date_time
        End Function

        ''' <summary>
        ''' フォルダの作成日時を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>作成日時</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetCreationTime(path_name As String) As DateTime
            Return GetTime(path_name, timeCreation, True)
        End Function

        ''' <summary>
        ''' フォルダの作成日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>作成日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetCreationTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeCreation, False)
        End Function

        ''' <summary>
        ''' フォルダの最後の書込日時を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>書込日時</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetLastWriteTime(path_name As String) As DateTime
            Return GetTime(path_name, timeLastWrite, True)
        End Function

        ''' <summary>
        ''' フォルダの最後の書込日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>書込日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetLastWriteTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeLastWrite, False)
        End Function

        ''' <summary>
        ''' フォルダの最後のアクセス日時を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>アクセス日時</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetLastAccessTime(path_name As String) As DateTime
            Return GetTime(path_name, timeLastAccess, True)
        End Function

        ''' <summary>
        ''' フォルダの最後のアクセス日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">フォルダ</param>
        ''' <returns>アクセス日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Shared Function GetLastAccessTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeLastAccess, False)
        End Function

        ''' <summary>
        ''' フォルダファイルコピー
        ''' </summary>
        ''' <param name="sourceFolderName">コピー元フォルダ</param>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <returns>true で成功</returns>
        Public Shared Function Copy(sourceFolderName As String, destFolderName As String) As Boolean
            Return Copy(sourceFolderName, destFolderName, 0)
        End Function

        ''' <summary>
        ''' フォルダファイルコピー
        ''' </summary>
        ''' <param name="sourceFolderName">コピー元フォルダ</param>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー設定</param>
        ''' <returns>true で成功</returns>
        Public Shared Function Copy(sourceFolderName As String, destFolderName As String, flag As Integer) As Boolean
            CheckInvalidPathName(sourceFolderName, True)
            CheckInvalidPathName(destFolderName, False)

            If NMtpCopyFolderFirst(sourceFolderName, destFolderName, flag Or CopyAllNoResult) = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピーファイルの取得</para>
        ''' </summary>
        ''' <param name="sourceFolderName">コピー元フォルダ</param>
        ''' <param name="destFolderName">コピー先フォルダ</param>
        ''' <param name="flag">コピー動作設定</param>
        ''' <exception cref="ArgumentNullException">
        '''	フォルダ が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	フォルダ がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	フォルダにアクセスできない
        ''' </exception>
        Public Sub CopyFirst(sourceFolderName As String, destFolderName As String, flag As Integer)
            CheckInvalidPathName(sourceFolderName, True)
            CheckInvalidPathName(destFolderName, False)

            _handle = NMtpCopyFolderFirst(sourceFolderName, destFolderName, flag)
            If _handle = CType(INVALID_HANDLE_VALUE, IntPtr) Then
                _handle = IntPtr.Zero
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>コピー実行</para>
        ''' </summary>
        ''' <param name="sourceFileName">コピー元のファイル</param>
        ''' <param name="destFileName">コピー先のファイル</param>
        ''' <param name="result">コピー結果</param>
        ''' <param name="error_code">コピー失敗時のエラーコード</param>
        ''' <returns>true でコピーファイル有</returns>
        Public Function CopyNext(ByRef sourceFileName As String, ByRef destFileName As String, ByRef result As Integer, ByRef error_code As Integer) As Boolean
            sourceFileName = ""
            destFileName = ""
            result = 0
            error_code = 0
            If _handle <> IntPtr.Zero Then
                Dim src As New StringBuilder(MAX_PATH)
                Dim dst As New StringBuilder(MAX_PATH)
                If NMtpCopyFolderNext(_handle, src, MAX_PATH, dst, MAX_PATH, result, _
                 error_code) Then
                    sourceFileName = src.ToString()
                    destFileName = dst.ToString()
                    Return True
                End If
            End If
            Return False
        End Function

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>終了</para>
        ''' </summary>
        Public Sub CopyClose()
            If _handle <> IntPtr.Zero Then
                NMtpCopyFolderClose(_handle)
                _handle = IntPtr.Zero
            End If
        End Sub

        ''' <summary>
        ''' <para>フォルダファイルコピー</para>
        ''' <para>エラー文字列取得</para>
        ''' </summary>
        Public Shared Function GetCopyErrorString(error_code As Integer) As String
            If error_code = ERROR_PATH_NOT_FOUND Then
                Return "Path not found"
            ElseIf error_code = ERROR_FILE_NOT_FOUND Then
                Return "File not found"
            ElseIf error_code = ERROR_FILE_EXISTS Then
                Return "File Exists"
            End If
            Return error_code.ToString()
        End Function

        ''' <summary>
        ''' フォルダファイルコピーのファイル数
        ''' </summary>
        Public ReadOnly Property CopyCount() As Integer
            Get
                If _handle = IntPtr.Zero Then
                    Return 0
                Else
                    Return NMtpCopyFolderCount(_handle)
                End If
            End Get
        End Property

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpCopyFile(src As String, dst As String, fail_exists As Integer) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpMoveFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpMoveFile(src As String, dst As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpDeleteFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpDeleteFile(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCreateFolder", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpCreateFolder(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpDeleteFolder", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpDeleteFolder(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpRemoveFolder", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpRemoveFolder(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindFirstFile", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpFindFirstFile(file_name As String, ByRef find_data As WIN32_FIND_DATA) As IntPtr
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindNextFile", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpFindNextFile(handle As IntPtr, ByRef find_data As WIN32_FIND_DATA) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindClose")> _
        Protected Shared Function NMtpFindClose(handle As IntPtr) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindDeviceFirst")> _
        Protected Shared Function NMtpFindDeviceFirst() As IntPtr
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindDeviceNext", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpFindDeviceNext(handle As IntPtr, name As StringBuilder, length As Integer) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpFindDeviceClose")> _
        Protected Shared Function NMtpFindDeviceClose(handle As IntPtr) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpGetStatus", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpGetStatus(path_name As String, ByRef find_data As WIN32_FIND_DATA) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderFirst", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpCopyFolderFirst(src As String, dst As String, flag As Integer) As IntPtr
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderCount")> _
        Protected Shared Function NMtpCopyFolderCount(handle As IntPtr) As Integer
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderNext", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpCopyFolderNext(handle As IntPtr, src_name As StringBuilder, src_length As Integer, dst_name As StringBuilder, dst_length As Integer, ByRef result As Integer, _
   ByRef error_code As Integer) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFolderClose")> _
        Protected Shared Function NMtpCopyFolderClose(handle As IntPtr) As Integer
        End Function
    End Class

    ''' <summary>
    ''' ファイル操作クラス
    ''' </summary>
    Public Class File
        ''' <summary>
        ''' <para>コピー先にファイルがあれば失敗</para>
        ''' <para>Copy の flag で指定</para>
        ''' </summary>
        Public Const CopyFailExists As Integer = &H1
        ''' <summary>
        ''' <para>新しいファイルのみコピー</para>
        ''' <para>Copy の flag で指定</para>
        ''' </summary>
        Public Const CopyNewer As Integer = &H10

        Private Const ERROR_FILE_NOT_FOUND As Integer = 2
        Private Const ERROR_PATH_NOT_FOUND As Integer = 3
        Private Const ERROR_ACCESS_DENIED As Integer = 5
        Private Const ERROR_FILE_EXISTS As Integer = 80
        Private Const ERROR_CANNOT_MAKE As Integer = 82

        Private Const MAX_PATH As Integer = 260

        Private Const FILE_ATTRIBUTE_DIRECTORY As UInteger = &H10

        ''' <summary>
        ''' <c>File</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
        End Sub

        ''' <summary>
        ''' ファイルコピー
        ''' </summary>
        ''' <param name="sourceFileName">コピー元ファイル</param>
        ''' <param name="destFileName">コピー先ファイル</param>
        ''' <exception cref="ArgumentNullException">
        '''	ファイルパス が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	ファイルパス がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="FileNotFoundException">
        '''	指定されたファイルが存在しない
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Sub Copy(sourceFileName As String, destFileName As String)
            Copy(sourceFileName, destFileName, True)
        End Sub

        ''' <summary>
        ''' ファイルコピー
        ''' </summary>
        ''' <param name="sourceFileName">コピー元ファイル</param>
        ''' <param name="destFileName">コピー先ファイル</param>
        ''' <param name="overwrite">true で上書き</param>
        ''' <exception cref="ArgumentNullException">
        '''	ファイルパス が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        ''' ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	ファイルパス がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="FileNotFoundException">
        '''	指定されたファイルが存在しない
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Sub Copy(sourceFileName As String, destFileName As String, overwrite As Boolean)
            Copy(sourceFileName, destFileName, If((overwrite), 0, CopyFailExists))
        End Sub

        ''' <summary>
        ''' ファイルコピー
        ''' </summary>
        ''' <param name="sourceFileName">コピー元ファイル</param>
        ''' <param name="destFileName">コピー先ファイル</param>
        ''' <param name="flag">コピー動作設定</param>
        ''' <exception cref="ArgumentNullException">
        '''	ファイルパス が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	ファイルパス がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="FileNotFoundException">
        '''	指定されたファイルが存在しない
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Sub Copy(sourceFileName As String, destFileName As String, flag As Integer)
            Directory.CheckInvalidPathName(sourceFileName, False)
            Directory.CheckInvalidPathName(destFileName, False)

            If Not NMtpCopyFile(sourceFileName, destFileName, flag) Then
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_FILE_NOT_FOUND Then
                    Throw New FileNotFoundException(message)
                ElseIf error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' ファイル移動
        ''' </summary>
        ''' <param name="sourceFileName">移動元ファイル</param>
        ''' <param name="destFileName">移動先ファイル</param>
        ''' <exception cref="ArgumentNullException">
        '''	ファイルパス が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	ファイルパス がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="FileNotFoundException">
        '''	指定されたファイルが存在しない
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Sub Move(sourceFileName As String, destFileName As String)
            Directory.CheckInvalidPathName(sourceFileName, False)
            Directory.CheckInvalidPathName(destFileName, False)

            If Not NMtpMoveFile(sourceFileName, destFileName) Then
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_FILE_NOT_FOUND Then
                    Throw New FileNotFoundException(message)
                ElseIf error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' ファイル削除
        ''' </summary>
        ''' <param name="path_name">削除するファイル</param>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="FileNotFoundException">
        '''	指定されたファイルが存在しない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Sub Delete(path_name As String)
            Directory.CheckInvalidPathName(path_name, False)

            If Not NMtpDeleteFile(path_name) Then
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_FILE_NOT_FOUND Then
                    Throw New FileNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
        End Sub

        ''' <summary>
        ''' ファイルの存在確認
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>true で存在</returns>
        Public Shared Function Exists(path_name As String) As Boolean
            Directory.CheckInvalidPathName(path_name, False)

            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()

            If NMtpGetStatus(path_name, find_data) Then
                If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = 0 Then
                    Return True
                End If
            End If
            Return False
        End Function

        Private Const timeCreation As Integer = 0
        Private Const timeLastWrite As Integer = 1
        Private Const timeLastAccess As Integer = 2

        Private Shared Function GetTime(path_name As String, time_type As Integer, local_flag As Boolean) As DateTime
            Directory.CheckInvalidPathName(path_name, False)

            Dim find_data As WIN32_FIND_DATA = New WIN32_FIND_DATA()
            Dim date_time As New DateTime()

            If NMtpGetStatus(path_name, find_data) Then
                If (find_data.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY) = 0 Then
                    If time_type = timeLastWrite Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftLastWriteTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastWriteTime.dwLowDateTime.ToString("X"), 16))
                    ElseIf time_type = timeCreation Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftCreationTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftCreationTime.dwLowDateTime.ToString("X"), 16))
                    ElseIf time_type = timeLastAccess Then
                        date_time = DateTime.FromFileTime((CLng(find_data.ftLastAccessTime.dwHighDateTime) << 32) + Convert.ToUInt32(find_data.ftLastAccessTime.dwLowDateTime.ToString("X"), 16))
                    End If
                End If
            Else
                Dim error_code As Integer = Marshal.GetLastWin32Error()
                Dim message As String = New System.ComponentModel.Win32Exception(error_code).Message
                If error_code = ERROR_PATH_NOT_FOUND Then
                    Throw New DirectoryNotFoundException(message)
                Else
                    Throw New IOException(message)
                End If
            End If
            If local_flag Then
                date_time.ToLocalTime()
            End If
            Return date_time
        End Function

        ''' <summary>
        ''' ファイルの作成日時を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>作成日時</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetCreationTime(path_name As String) As DateTime
            Return GetTime(path_name, timeCreation, True)
        End Function

        ''' <summary>
        ''' ファイルの作成日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>作成日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetCreationTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeCreation, False)
        End Function

        ''' <summary>
        ''' ファイルの最後の書込日時を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>書込日時</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetLastWriteTime(path_name As String) As DateTime
            Return GetTime(path_name, timeLastWrite, True)
        End Function

        ''' <summary>
        ''' ファイルの最後の書込日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>書込日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetLastWriteTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeLastWrite, False)
        End Function

        ''' <summary>
        ''' ファイルの最後のアクセス日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>アクセス日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetLastAccessTime(path_name As String) As DateTime
            Return GetTime(path_name, timeLastAccess, True)
        End Function

        ''' <summary>
        ''' ファイルの最後のアクセス日時(世界協定時刻)を取得
        ''' </summary>
        ''' <param name="path_name">ファイル</param>
        ''' <returns>アクセス日時(世界協定時刻)</returns>
        ''' <exception cref="ArgumentNullException">
        '''	path が null
        '''	</exception>
        '''	<exception cref="ArgumentException">
        '''	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
        ''' </exception>
        '''	<exception cref="PathTooLongException">
        '''	path がシステム定義の最大値を超えている
        ''' </exception>
        '''	<exception cref="DirectoryNotFoundException">
        '''	指定されたパスが正しくない
        ''' </exception>
        '''	<exception cref="IOException">
        '''	ファイルにアクセスできない
        ''' </exception>
        Public Shared Function GetLastAccessTimeUtc(path_name As String) As DateTime
            Return GetTime(path_name, timeLastAccess, False)
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpCopyFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpCopyFile(src As String, dst As String, fail_exists As Integer) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpMoveFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpMoveFile(src As String, dst As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpDeleteFile", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpDeleteFile(path_name As String) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpGetStatus", CharSet:=CharSet.Unicode, SetLastError:=True)> _
        Protected Shared Function NMtpGetStatus(path_name As String, ByRef find_data As WIN32_FIND_DATA) As Boolean
        End Function
    End Class


    ''' <summary>
    ''' ファイルを開くダイアログ表示クラス
    ''' </summary>
    Public Class FileDialog
        ''' <summary>
        ''' 選択したデバイスパスの先頭の PC/コンピュータ を削除
        ''' </summary>
        Protected Const DeleteTopPC As UInteger = &H1
        ''' <summary>
        ''' SaveDialog でファイルの上書き確認を行う
        ''' </summary>
        Protected Const SaveOverWritePrompt As UInteger = &H2
        ''' <summary>
        ''' SaveDialog で指定された拡張子のファイルのみ選択可能
        ''' </summary>
        Protected Const StrictFileTypes As UInteger = &H4
        ''' <summary>
        ''' カレントフォルダを変更しない
        ''' </summary>
        Protected Const NoChangeDir As UInteger = &H8
        ''' <summary>
        ''' 複数項目選択可能
        ''' </summary>
        Protected Const AllowMultiSelect As UInteger = &H200
        ''' <summary>
        ''' フォルダの存在が必要
        ''' </summary>
        Protected Const PathMustExist As UInteger = &H800
        ''' <summary>
        ''' SaveDialog でファイルの存在が必要
        ''' </summary>
        Protected Const FileMustExist As UInteger = &H1000
        ''' <summary>
        ''' SaveDialog で存在しない場合通知
        ''' </summary>
        Protected Const SaveCreatePrompt As UInteger = &H2000
        ''' <summary>
        ''' SaveDialog で読み取り専用不可
        ''' </summary>
        Protected Const NoReadOnlyReturn As UInteger = &H8000
        ''' <summary>
        ''' 最近開いたアイテムや保存した場所を非表示
        ''' </summary>
        Protected Const HideMruPlace As UInteger = &H20000
        ''' <summary>
        ''' ナビゲーションペインを非表示
        ''' </summary>
        Protected Const HidePinnedPlaces As UInteger = &H40000
        ''' <summary>
        ''' ショートカットを対象としない
        ''' </summary>
        Protected Const NoDereferenceLinks As UInteger = &H100000
        ''' <summary>
        ''' 最新のドキュメントに開いたファイルを追加しない
        ''' </summary>
        Protected Const DontAddToRecent As UInteger = &H2000000
        ''' <summary>
        ''' 隠しファイルとシステムアイテムを含む
        ''' </summary>
        Protected Const ShowHidden As UInteger = &H10000000
        ''' <summary>
        ''' OpenFile でプレビューを表示する
        ''' </summary>
        Protected Const PreviewPanelOn As UInteger = &H40000000

        Protected Const MAX_PATH As Integer = 260

        ''' <summary>
        ''' <c>FileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
            Handle = IntPtr.Zero
            CheckPathExists = True
            CheckFileExists = True
            RestoreDirectory = True
            FilterIndex = 1
        End Sub
        ''' <summary>
        ''' <c>FileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="hwnd">親ウインドウハンドル</param>
        Public Sub New(hwnd As IntPtr)
            Handle = hwnd
            CheckPathExists = True
            CheckFileExists = True
            RestoreDirectory = True
            FilterIndex = 1
        End Sub

        Protected Overridable Function MakeFolderOption() As UInteger
            Dim folder_option As UInteger = 0
            If CheckFileExists Then
                folder_option = folder_option Or FileMustExist
            End If
            If CheckPathExists Then
                folder_option = folder_option Or PathMustExist
            End If
            If RestoreDirectory Then
                folder_option = folder_option Or NoChangeDir
            End If
            Return folder_option
        End Function

        ''' <summary>
        ''' 既定の拡張子を取得または設定
        ''' </summary>
        Public Property DefaultExt() As String
            Get
                Return m_DefaultExt
            End Get
            Set(value As String)
                m_DefaultExt = Value
            End Set
        End Property
        Private m_DefaultExt As String
        ''' <summary>
        ''' 既定のファイル名を設定、選択されたファイル名を取得
        ''' </summary>
        Public Property FileName() As String
            Get
                Return m_FileName
            End Get
            Set(value As String)
                m_FileName = Value
            End Set
        End Property
        Private m_FileName As String
        ''' <summary>
        ''' フィルター文字列の取得または設定
        ''' </summary>
        Public Property Filter() As String
            Get
                Return m_Filter
            End Get
            Set(value As String)
                m_Filter = Value
            End Set
        End Property
        Private m_Filter As String
        ''' <summary>
        ''' 既定フィルター番号の取得又は設定
        ''' </summary>
        Public Property FilterIndex() As Integer
            Get
                Return m_FilterIndex
            End Get
            Set(value As Integer)
                m_FilterIndex = Value
            End Set
        End Property
        Private m_FilterIndex As Integer
        ''' <summary>
        ''' <para>既定フォルダー名の取得または設定</para>
        ''' <para>※MTP デバイス上のパスは指定できません</para>
        ''' </summary>
        Public Property InitialDirectory() As String
            Get
                Return m_InitialDirectory
            End Get
            Set(value As String)
                m_InitialDirectory = Value
            End Set
        End Property
        Private m_InitialDirectory As String
        ''' <summary>
        ''' ダイアログタイトルの取得または設定
        ''' </summary>
        Public Property Title() As String
            Get
                Return m_Title
            End Get
            Set(value As String)
                m_Title = Value
            End Set
        End Property
        Private m_Title As String
        ''' <summary>
        ''' 親ウインドウハンドルの取得または設定
        ''' </summary>
        Public Property Handle() As IntPtr
            Get
                Return m_Handle
            End Get
            Set(value As IntPtr)
                m_Handle = Value
            End Set
        End Property
        Private m_Handle As IntPtr
        ''' <summary>
        ''' 存在しないファイルの名前をユーザーが指定した場合に警告を表示するかどうかを示す値を取得または設定
        ''' </summary>
        Public Property CheckFileExists() As Boolean
            Get
                Return m_CheckFileExists
            End Get
            Set(value As Boolean)
                m_CheckFileExists = Value
            End Set
        End Property
        Private m_CheckFileExists As Boolean
        ''' <summary>
        ''' 存在しないパスをユーザーが指定した場合に警告を表示するかどうかを示す値を取得または設定
        ''' </summary>
        Public Property CheckPathExists() As Boolean
            Get
                Return m_CheckPathExists
            End Get
            Set(value As Boolean)
                m_CheckPathExists = Value
            End Set
        End Property
        Private m_CheckPathExists As Boolean
        ''' <summary>
        ''' カレントフォルダを変更するかどうかを示す値を取得または設定
        ''' </summary>
        Public Property RestoreDirectory() As Boolean
            Get
                Return m_RestoreDirectory
            End Get
            Set(value As Boolean)
                m_RestoreDirectory = Value
            End Set
        End Property
        Private m_RestoreDirectory As Boolean
    End Class

    ''' <summary>
    ''' ファイルを開くダイアログ表示クラス
    ''' </summary>
    Public Class OpenFileDialog
        Inherits FileDialog
        Protected _file_list As New List(Of String)()

        ''' <summary>
        ''' <c>OpenFileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
            MyBase.New()
            AllowSelect = False
            PreviewPanel = False
        End Sub
        ''' <summary>
        ''' <c>OpenFileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="hwnd">親ウインドウハンドル</param>
        Public Sub New(hwnd As IntPtr)
            MyBase.New(hwnd)
            AllowSelect = False
            PreviewPanel = False
        End Sub

        ''' <summary>
        ''' 複数ファイルを選択可能かどうかを示す値を取得または設定
        ''' </summary>
        Public Property AllowSelect() As Boolean
            Get
                Return m_AllowSelect
            End Get
            Set(value As Boolean)
                m_AllowSelect = Value
            End Set
        End Property
        Private m_AllowSelect As Boolean
        ''' <summary>
        ''' プレビュー表示をするかどうかを示す値を取得または設定
        ''' </summary>
        Public Property PreviewPanel() As Boolean
            Get
                Return m_PreviewPanel
            End Get
            Set(value As Boolean)
                m_PreviewPanel = Value
            End Set
        End Property
        Private m_PreviewPanel As Boolean
        ''' <summary>
        ''' 選択された全てのファイル名を取得
        ''' </summary>
        Public ReadOnly Property FileNames() As String()
            Get
                Return _file_list.ToArray()
            End Get
        End Property

        Protected Overrides Function MakeFolderOption() As UInteger
            Dim folder_option As UInteger = MyBase.MakeFolderOption()
            If AllowSelect Then
                folder_option = folder_option Or AllowMultiSelect
            End If
            If PreviewPanel Then
                folder_option = folder_option Or PreviewPanelOn
            End If
            Return folder_option
        End Function

        ''' <summary>
        ''' ファイルを開くダイアログを表示
        ''' </summary>
        ''' <returns>DialogResult.OK でファイル選択</returns>
        Public Function ShowDialog() As System.Windows.Forms.DialogResult
            Dim sname As New StringBuilder(MAX_PATH)
            Dim result As Integer
            Dim folder_option As UInteger = MakeFolderOption()
            result = NMtpOpenDialog(Handle, Title, InitialDirectory, FileName, Filter, FilterIndex, _
             DefaultExt, sname, MAX_PATH, folder_option)
            If result <> 0 Then
                FileName = sname.ToString()
                If (folder_option And AllowMultiSelect) = AllowMultiSelect Then
                    _file_list = New List(Of String)()
                    _file_list.Add(FileName)
                    While NMtpOpenDialogNext(result, sname, MAX_PATH)
                        _file_list.Add(sname.ToString())
                    End While
                    NMtpOpenDialogClose(result)
                End If
                Return System.Windows.Forms.DialogResult.OK
            End If
            Return System.Windows.Forms.DialogResult.Cancel
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpOpenDialog", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpOpenDialog(wnd As IntPtr, title As String, folder As String, name As String, filter As String, index As Integer, _
   ext As String, filename As StringBuilder, length As UInteger, flag As UInteger) As Integer
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpOpenDialogNext", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpOpenDialogNext(handle As Integer, filename As StringBuilder, length As UInteger) As Boolean
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpOpenDialogClose", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpOpenDialogClose(handle As Integer) As Boolean
        End Function
    End Class

    ''' <summary>
    ''' ファイル保存ダイアログ表示クラス
    ''' </summary>
    Public Class SaveFileDialog
        Inherits FileDialog
        ''' <summary>
        ''' <c>SaveFileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
            MyBase.New()
            CreatePrompt = True
        End Sub
        ''' <summary>
        ''' <c>SaveFileDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="hwnd">親ウインドウハンドル</param>
        Public Sub New(hwnd As IntPtr)
            MyBase.New(hwnd)
            CreatePrompt = True
        End Sub

        ''' <summary>
        ''' 存在しないファイルが選択された場合に許可表示を行うかどうかを示す値を取得または設定
        ''' </summary>
        Public Property CreatePrompt() As Boolean
            Get
                Return m_CreatePrompt
            End Get
            Set(value As Boolean)
                m_CreatePrompt = Value
            End Set
        End Property
        Private m_CreatePrompt As Boolean

        Protected Overrides Function MakeFolderOption() As UInteger
            Dim folder_option As UInteger = MyBase.MakeFolderOption()
            If CreatePrompt Then
                folder_option = folder_option Or SaveCreatePrompt
            End If
            Return folder_option
        End Function
        ''' <summary>
        ''' ファイルを保存ダイアログを表示
        ''' </summary>
        ''' <returns>DialogResult.OK でファイル選択</returns>
        Public Function ShowDialog() As System.Windows.Forms.DialogResult
            Dim sname As New StringBuilder(MAX_PATH)
            If NMtpSaveDialog(Handle, Title, InitialDirectory, FileName, Filter, FilterIndex, _
             DefaultExt, sname, MAX_PATH, MakeFolderOption()) Then
                FileName = sname.ToString()
                Return System.Windows.Forms.DialogResult.OK
            End If
            Return System.Windows.Forms.DialogResult.Cancel
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpSaveDialog", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpSaveDialog(wnd As IntPtr, title As String, folder As String, name As String, filter As String, filter_index As Integer, _
   ext As String, filename As StringBuilder, length As Integer, flag As UInteger) As Boolean
        End Function
    End Class

    ''' <summary>
    ''' フォルダ選択ダイアログ表示クラス
    ''' </summary>
    Public Class FolderBrowserDialog
        Private Const MAX_DIR_PATH As Integer = 248

        ''' <summary>
        ''' <c>FolderBrowserDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        Public Sub New()
            Handle = IntPtr.Zero
        End Sub
        ''' <summary>
        ''' <c>FolderBrowserDialog</c>クラスの新規インスタンスを初期化します。
        ''' </summary>
        ''' <param name="hwnd">親ウインドウハンドル</param>
        Public Sub New(hwnd As IntPtr)
            Handle = hwnd
        End Sub

        ''' <summary>
        ''' 親ウインドウハンドルの取得または設定
        ''' </summary>
        Public Property Handle() As IntPtr
            Get
                Return m_Handle
            End Get
            Set(value As IntPtr)
                m_Handle = Value
            End Set
        End Property
        Private m_Handle As IntPtr
        ''' <summary>
        ''' ダイアログタイトルの取得または設定
        ''' </summary>
        Public Property Description() As String
            Get
                Return m_Description
            End Get
            Set(value As String)
                m_Description = Value
            End Set
        End Property
        Private m_Description As String
        ''' <summary>
        ''' <para>既定フォルダー名の取得または設定</para>
        ''' <para>※MTP デバイス上のパスは指定できません</para>
        ''' </summary>
        Public Property RootFolder() As String
            Get
                Return m_RootFolder
            End Get
            Set(value As String)
                m_RootFolder = Value
            End Set
        End Property
        Private m_RootFolder As String
        ''' <summary>
        ''' 選択されたフォルダ名を取得
        ''' </summary>
        Public Property SelectPath() As String
            Get
                Return m_SelectPath
            End Get
            Set(value As String)
                m_SelectPath = Value
            End Set
        End Property
        Private m_SelectPath As String

        ''' <summary>
        ''' フォルダ選択ダイアログを表示
        ''' </summary>
        ''' <returns>DialogResult.OK でフォルダ選択</returns>
        Public Function ShowDialog() As System.Windows.Forms.DialogResult
            Dim sname As New StringBuilder(MAX_DIR_PATH)
            If NMtpOpenFolderDialog(Handle, Description, RootFolder, sname, MAX_DIR_PATH, 0) Then
                SelectPath = sname.ToString()
                Return System.Windows.Forms.DialogResult.OK
            End If
            Return System.Windows.Forms.DialogResult.Cancel
        End Function

        <DllImport("nMtpUtil.DLL", EntryPoint:="NMtpOpenFolderDialog", CharSet:=CharSet.Unicode)> _
        Protected Shared Function NMtpOpenFolderDialog(wnd As IntPtr, title As String, folder As String, filename As StringBuilder, length As Integer, flag As Integer) As Boolean
        End Function
    End Class
End Namespace
